/***************************************************
  Proceed commands, such as:
        Enable joystick
        Show grid
        Change level
-----------------------------------------------------
        Linux version   07.04.98        M.G.
****************************************************/
#include <sys/time.h>
#include <pwd.h>        // For getting real name throgh getpwuid
#include <unistd.h>     // For getting user login name
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "tetX.h"
#include "egafont.h"
#include "keyjoy.h"

#ifdef X11
#define EGA -1
#else
#include <vga.h>
#endif

extern BOARD Board[BRD_HEIGHT][BRD_WIDTH];

extern FIGURE NexFig;
extern long  Ticks;                     // Timer ticks from program start
extern char *DemHeader, *DemText[];
extern char *PlayListFileName;
extern int  JoyEquiv[2];

extern int  Level, Layers;              // Current level and number of layers
extern int  NexTyp, NexRot;             // Next type and rotation
extern unsigned int  Score;             // Current score
extern int  Bonus;                      // Current bonus
extern int  VgaVideoMode;               // G...x...x...

extern int  JoyEnabled;                 // True if joystick is enabled
extern int  ScrollAvailable;            // True id scroller has to be shown
extern int  MusicEnabled;               // True if BG music is enabled
extern int  TimerSuspended;             // True if no timer interrupt trap
extern int  Demo;                       // True for demo game
extern int  ShowGrid;                   // True, if grid has to be shown
extern int  ShowNext;                   // True, if next figure has to be shown
extern char *VarTexts[];

extern unsigned int PrevScore;          // Score before start of current level
extern int  PrevLays;                   // Full layers before start of current level

extern int  NewLevelAt;                 // Score where new level arrives
extern int  DelayMove, DelayJoy;        // Delay before moving figure or
                                        // checking joystick

extern long AdvTicks;                   // Ticks when figure should be advanced
extern long AdvJoy;                     // Ticks when joystick should be checked

extern int ScreenHeight, ScreenWidth, ScreenDepth;
extern int BoardX, BoardY, BoardW, BoardH;
extern int ScoreX, ScoreY, ScoreW, ScoreH;
extern int JoyX, JoyY, JoyW, JoyH;
extern int NextX, NextY, NextW, NextH;
extern int ScrollX, ScrollH, ScrollY, ScrollW, ScrollEY, ShadowY;
extern int FgrColour, BkgrColour, TxtColour, HiliteColour;
extern int VideoChipset, ScreenBusy;

extern int FullLays;                    // Number of layers to be filled prior to game
extern int Moves;                       // Number of figures presented
extern int   DefLevel, DefLays;         // Starting level and layers
extern int   FigTurns[];
extern WHERE_WE_ARE where;              // PROLOGUE, BETWEEN, etc
extern BOTTOM_MODE  BottomMode;
extern IMAGE_HEADER owl256_Hdr;
extern char FillPats[][8];

//  Statistics
extern unsigned int  Scores[10];        // Scores for each level
extern int     Statist[NFIGS];          // Statistics by number of figures
extern int     StatLays[10];            // Layers for each level



// Tables
static int Delays[] = { 10, 9, 8, 7, 6, 5, 4, 3, 2, 1};
static int MovesOnLevel[] = {20, 30, 40, 50, 60, 70, 80, 90, 100};
//static int JoyDelays[] = { 3, 2, 2, 2, 2, 2, 1, 1, 0, 0};
static int JoyDelays[] = { 2, 2, 2, 2, 1, 1, 1, 0, 0, 0};
static BOARD brd[BRD_HEIGHT][BRD_WIDTH];

// 'Effective' coordinates
//  (i.e upper left corner of writing space)
int ScoreEX, ScoreEY;
int NextEX, NextEY;
int JoyEX, JoyEY;
int HelpPageNo, HelpPageQ;

// Static data
static int ItIsNewGame = 1;
static int RestartScroller = 1;
/************************************************************/
/*                   EXECUTABLE CODE                        */
/************************************************************/
void change_bcol(void) { ; }


//----------------------------------------------------
void PrepareShowGrid(FIGURE figure, int x, int y, int wid, int hei)
{
    int  i;

    for (i = 0; i<BRD_HEIGHT; i++)
         memcpy( brd[i],  Board[i], BRD_WIDTH*sizeof(BOARD));

    embed_figure(brd, figure, x, y, wid, hei);

    DrawBoard(brd, 0, BRD_WIDTH, 0, BRD_HEIGHT);

    if (Demo)
        drawrobbrd(brd,  0, BRD_WIDTH, 0, BRD_HEIGHT);
}

//----------------------------------------------------
int PrepareMusic(int Param)
{  int rc = 0;
                       /* PROLOGUE   BETWEEN           RESULTS    ADVICE  */
   static int Ind[6] =  {   0,          1,                12,        13};
   static int Mode[6] = {   0,  MODE_FORCE|MODE_CHECK,     0,         0};



   if (MusicEnabled)
   {  if(where == W_GAME)
           rc = PlayPlayLine(Level+2, Param);
      else
           rc = PlayPlayLine(Ind[where], Mode[where] | Param);

      return rc;

   }
   else
     StopBGPlay(0);

   return rc;

}
//----------------------------------------------------
void SetLevel(void)
{
  long  t;


   PrevScore = Score;
   PrevLays = Layers;

   NewLevelAt += MovesOnLevel[Level];
   DelayMove = Delays[Level];
   if (!Demo) DelayMove++;
   DelayJoy  = JoyDelays[Level];

   PrepareMusic(Level == DefLevel ? MODE_FORCE : 0);

   t = Ticks;
   AdvTicks = t + DelayMove;
   AdvJoy = t + DelayJoy;

   joyGetStatus();    /* This will clean joystick queue */

#ifndef X11
   while (vga_getkey() != 0);	/* Clearing keyboard buffer */
#endif
   	
}


void ChangeLevel(FIGURE figure, int x, int y, int wid, int hei)
{
   int  i, j;
   int  sw = 0;
   long NextTicks;

   Scores[Level] = (Score > PrevScore) ? (Score - PrevScore) : 0;
   StatLays[Level] = Layers - PrevLays;

   PrevScore = Score;   // Dupplicate, but may fix a strange bug
   PrevLays = Layers;

   if (VideoChipset == EGA)
   {  AllocScreenContext();
      ScreenToScreenContext();
   }

   for (i=0; i<2 || sw; i++)
   {
       for (j=0; j<2; j++)   // Was 2
           {   NextTicks = Ticks + 2;
               if (VideoChipset == EGA)
               {  InvertScreenContext();
                  ScreenContextToScreen();
               }
               else
                  InvertPalette();

	       sleep_until(NextTicks);	
           }


    }

   if (VideoChipset == EGA)
   {     ScreenContextToScreen();
         DeleteScreenContext();
   }


   Level++;

   SetLevel();

   if (x>= 0)
   {
      int  i;

      for (i = 0; i<BRD_HEIGHT; i++)
           memcpy( brd[i],  Board[i], BRD_WIDTH*sizeof(BOARD));

      embed_figure(brd, figure, x, y, wid, hei);
      DrawAll(brd);
    }
    else
     DrawAll(Board);

}

//----------------------------------------------------
void NewGame(void)
{   int i;
    struct timeval  tv;
    struct timezone tz;


    gettimeofday(&tv, &tz);
    srand((unsigned int) tv.tv_usec+tv.tv_sec);

    Level = DefLevel;
    Layers = 0;
    Score = 0;
    Moves = 0;
    FullLays = DefLays;

    for (i=0; i<7; i++) Statist[i] = 0;
    for (i=0; i<Level; i++)  Scores[i] = StatLays[i] = 0;

    for (i = 0; i< BRD_HEIGHT; i++)     /* Initialize playing board; */
        memset(Board[i], 0, BRD_WIDTH*sizeof(BOARD));

    if (FullLays > 0) FillLayers(FullLays);

    NexTyp = get_rand(NFIGS);
    NexRot = get_rand(FigTurns[NexTyp]);

    get_figure(NexFig, NexTyp, NexRot);
    ShowNext = 0;
    ShowGrid = 0;

    PrevLays = PrevScore = 0;
    NewLevelAt = 0;

    InitHelpPages();

    if (ItIsNewGame && Demo)
    {  BottomMode = START_BOTTOM_DEMO;
       RestartScroller = 1;
    }
    else if (!Demo)
    {  BottomMode = START_BOTTOM_GAME;
#ifdef SHOW_256C_IMAGE
      setPalette(&owl256_Hdr, 1 << ScreenDepth);
#endif
    }

    where = W_GAME;

    SetLevel();
    ItIsNewGame = 0;

    DrawAll(Board);

}


// par = 0 - continue without pause
//       4 - continue with pause
int EndGame(int par)
{
  int  rc;


// Flush keyboard buffer
  while (kbdGet()!=0);

  if (Demo) robthink(3);

  Scores[Level] = (Score > PrevScore) ? (Score - PrevScore) : 0;
  StatLays[Level] = Layers - PrevLays;

  rc = ShowStatistics();
  if (par==0 || rc!=0 ) return rc;


  StopScroller();

  ItIsNewGame = 1;
  return ShowIntergameScreen();

}

//----------------------------------------------------

void ShowStatElement(int row, unsigned int value)
{
   int h = egaGetFontHeight() + ROW_DISTANCE;
   int y = ScoreEY + h*row;
   int x;
   char buffer[11];

   sprintf(buffer, "%u", value);


#ifdef RIGHT_ADJUSTMENT
    x = ScoreEX;
#else
    x = ((TABLE_WIDTH-strlen(buffer))<<3) + ScoreEX;
#endif

    egaShowText(buffer, x, y, TxtColour, BkgrColour);

}

void PrepareStatWindow(void)
{

   int  values[4];

   char *text;

   int i, h, x, y;
   RECT  rc;

   rc.top = ScoreY;
   rc.left = ScoreX;

   rc.bottom = ScoreY + ScoreH;
   rc.right = ScoreX + ScoreW;

   values[0] = Score;
   values[1] = Level;
   values[2] = Layers;
   values[3] = Bonus;

   DrawFramedRect(&rc);

   h = egaGetFontHeight();
   y = h*4 + ROW_DISTANCE*3;
   ScoreEY = y = ((ScoreH - y) >> 1) + ScoreY;
   ScoreEX = ScoreX + 8;

   h +=ROW_DISTANCE;

   for (i=0; i<4; i++)
   {
      text = VarTexts[i];

#ifdef RIGHT_ADJUSTMENT
      x = ScoreEX + (TABLE_WIDTH-strlen(text))*8;
#else
      x = ScoreEX;
#endif
      egaShowText(text, x, y, TxtColour, BkgrColour);

      ShowStatElement(i, values[i]);
      y += h;
   }


}

#define BONUS_ROW 3
void ShowBonus(void)
{
    ShowStatElement(BONUS_ROW, Bonus);
}

void HideBonus(void)
{  int h = egaGetFontHeight();
   int y = ScoreEY + (h + ROW_DISTANCE) * BONUS_ROW;
   RECT rc;

   rc.top = y; rc.bottom = y+h;
   rc.left = ScoreX; rc.right = ScoreX + ScoreW;
   DrawSolidRect(&rc, BkgrColour);

}

void UpdateStatus(void)
{
    ShowStatElement(0, Score);
    ShowStatElement(2, Layers);

}

//-----------------------------------------------------------
#define JOY_ARROW_WIDTH  79
#define JOY_ARROW_HEIGHT 64

void PrepareJoyWindow(void)

{
   RECT rc;
   char *joytxt = VarTexts[JoyTextPos];


   rc.top = JoyY;
   rc.left = JoyX;

   rc.bottom = JoyY + JoyH;
   rc.right = JoyX + JoyW;

   if (JoyEnabled || Demo)
   {
     JoyEX =  ((JoyW - egaGetTextWidth(joytxt)) >> 1) + JoyX;
     JoyEY = JoyY + 3;

     DrawFramedRect(&rc);

     egaShowText(joytxt, JoyEX, JoyEY, TxtColour, BkgrColour);

     JoyEY =  JoyY + JoyH - JOY_ARROW_HEIGHT;
     JoyEX =  JoyX + ((JoyW - JOY_ARROW_WIDTH) >> 1);

   }
   else
   { rc.top -= 8 + (BoardY%8);
     rc.bottom += 8;
     rc.left -= 8;
     rc.right += 8;
     DrawBkgrPattern(&rc);
   }



}


//-----------------------------------------------------------

void ShowNextFigure(void)
{
 // show_figure(NexFig, NextEX, NextEY, FIG_DIM, FIG_DIM);
    ShowFigure(NexFig, NextEX, NextEY, FIG_DIM, FIG_DIM);
   
}


void PrepareNextWindow(void)
{  RECT rc;
   int  FigFWid, FigFHei;
   char *nexttxt = VarTexts[NextTextPos];

   rc.top = NextY;
   rc.left = NextX;

   rc.bottom = NextY + NextH;
   rc.right = NextX + NextW;


   if (ShowNext)
   {
     DrawFramedRect(&rc);

     NextEX = NextX + ((NextW - egaGetTextWidth(nexttxt)) >> 1);
     NextEY = NextY + 3;
     egaShowText(nexttxt, NextEX, NextEY, TxtColour, BkgrColour);

     FigFWid = FIG_WIDTH * FIG_DIM;
     FigFHei = FIG_HEIGHT * FIG_DIM;

     NextEX = NextX + ((NextW - FigFWid) >> 1);
     NextEY = NextY + NextH - FigFHei - 3;

     ShowNextFigure();
   }
   else
   { rc.top -= 10 + (BoardY%8);
     rc.bottom += 10;
     rc.left -= 8;
     rc.right += 8;
     DrawBkgrPattern(&rc);
   }

}

//====================================================
// Random filling of layers
void FillLayers(int layers)
{   int i, x, y, sgn, fill;


    for (i=0; i<layers; i++)
    { y = BRD_HEIGHT - 1 - i;
      
     do 
     {  sgn = 1;
        for (x =0; x<BRD_WIDTH; x++)
        {  fill = get_rand1(16);
           if (fill == 0 || fill == 8 || fill >= 12 ||
             ( x == (BRD_WIDTH- 1) && sgn ) ) { fill = 0; sgn = 0;  }
           Board[y][x]=fill;
        }
     } while (sgn); 
    }
}

//====================================================
void ShowShadowFrame(void)
{   RECT rc;

    rc.top = ShadowY; rc.bottom = ShadowY + FIG_HEIGHT;
    rc.left = BoardX; rc.right = BoardX + BoardW;


    if (BottomMode == B_SHADOW && where == W_GAME)
            DrawFramedRect(&rc);
    else
    {  rc.top -= 5;  rc.bottom += 8;
       if (rc.bottom > ScreenHeight) rc.bottom = ScreenHeight;
       rc.left -=8;  rc.right += 8;
       DrawPatRect(&rc, FillPats[Level], 6-(BoardY%2));
    }
}


//====================================================
static long NextHelpPageAt;
//  static char JoyB1[2] = {'\0', '\0'};
//  static char JoyB2[2] = {'\0', '\0'}; 

static void (*ShowHelpPage)(void);

static char *Txt11[] =
    { " \x1B 4", "Left  ", "\x1A 6", "Right ",
      "\x19 2", "Down ", "N 1", "Next ",
      "\x18 5 Enter", "Rotate", NULL};

#ifdef X11
static char *Txt11A[] =
    { "  4 ", "Left  ", " 6 ", "Right ",
      " 2 ", "Down ", "N 1", "Next ",
      " 5 Enter", "Rotate", NULL};
#endif

static char *Txt12[] =
    { "  M ", "Music ", " L ", "Level ",
      " G ", "Grid ", " Q ", "Quit ",
      " 0 Space ", "Drop", NULL};

static char *Txt21[] =
   {  " J", "Enable / calibrate joystick",
      " K",  "Disable joystick - keys only",
      NULL
   };

static char *Txt22[] =
   {
      " B", "Bottom: Shadow/Help/None   ",
      " H", "Next Help Page",
      "    P ", "Pause",  NULL
   };

//static char *Txt31[] =
//   { "Joystick: Left, Right", "Move aside",
//         "Up", "Rotate", "Down", "Move down", NULL};

//static char *Txt32[] =
//   { "Joystick button equivalents: Button1 =", JoyB1,
//                                  "Button2 = ", JoyB2, NULL};


static char *TxtD1[] =
    { " M", "Music",  "P", "Pause",
      "L", "Level", "G", "Grid", "Q", "Quit",
      "B", "Scroller/Help/None", NULL};

static char **PageGameVGA[][2] = {{Txt11, Txt12}, {Txt21, Txt22}}; //{Txt31, Txt32}};
static char **PageDemoVGA[][2]  = {{TxtD1, NULL} };
#ifndef X11
static char **PageGameEGA[] = {Txt11, Txt12, Txt21, Txt22}; //, Txt31, Txt32};
static char **PageDemoEGA[] = {TxtD1, NULL};
#endif

static char  HelpBusy = 0;

#define HELP_INTERVAL 182

void  ShowHelpPageVGA(void)
{
   char  **rowptr, *txtptr;
   int row, col, x, y, fh;

   y = ScrollEY;
   if (Demo) y += 8;
   fh = egaGetFontHeight();

   for (row = 0; row <2; row++)
   {
      rowptr = Demo ? PageDemoVGA[HelpPageNo][row]
                    : PageGameVGA[HelpPageNo][row];
      if (!rowptr) continue;
      x = ScrollX + 8;

      for (col = 0; col < 6; col++)
      {
        txtptr = *rowptr++;
        if (!txtptr) break;

        egaShowText(txtptr, x, y, HiliteColour, BkgrColour);
        x += egaGetTextWidth(txtptr) + 8;

        txtptr = *rowptr++;
        egaShowText(txtptr, x, y, TxtColour, BkgrColour);
        x += egaGetTextWidth(txtptr) + 16;


      }
      y += fh;

   }
}

#ifndef X11
void  ShowHelpPageEGA(void)
{
   char  **rowptr, *txtptr;
   int  col, x, y, fh;

   y = ScrollEY;
   fh = egaGetFontHeight();

   rowptr = Demo ? PageDemoEGA[HelpPageNo]
                 : PageGameEGA[HelpPageNo];

   x = ScrollX + 8;

   for (col = 0; col < 6; col++)
   {
      txtptr = *rowptr++;
      if (!txtptr) break;

      egaShowText(txtptr, x, y, HiliteColour, BkgrColour);
      x += egaGetTextWidth(txtptr) + 8;

      txtptr = *rowptr++;
      egaShowText(txtptr, x, y, TxtColour, BkgrColour);
      x += egaGetTextWidth(txtptr) + 16;
   }

}
#endif

void   PrepareScroller(void)
{
#ifndef X11 
    if ((BottomMode == B_SCROLLER) && Demo && (!ScrollAvailable))
                               BottomMode ++;
#endif

    switch(BottomMode)
    { case  B_SCROLLER:
#ifndef X11
         if (Demo && RestartScroller)
         {

#if !defined(SCROLLER_DISABLED)  // && !defined(X11) 
	     int n = getDemTextSize();
             StartScroller(DemHeader, 1, n, DemText);
#endif

             RestartScroller = 0;
         }
         else
         if (Demo) ShowScrollerFrame();
         else
#endif
          ShowShadowFrame();

         break;

      case  B_HELP:
         StopScroller();
         ShowScrollerFrame();
         InitHelpPages();
         (*ShowHelpPage)();
         break;


/*    case  B_NONE:	*/
      default:
         break;

	

    }

}

void InitHelpPages(void)
{
    HelpPageNo = 0;
    HelpPageQ = 2;
//    JoyB1[0] = JoyEquiv[0];
//    JoyB2[0] = JoyEquiv[1];
    NextHelpPageAt = Ticks + HELP_INTERVAL;

#ifndef X11
    if (ScreenHeight > 350)
    {
#else
      if (egaFontType() == FONT_TYPE_X11_NATURAL)
           PageGameVGA[0][0] = Txt11A;
#endif

       ShowHelpPage = ShowHelpPageVGA;
       HelpPageQ =  Demo ? 1 : 2 ;   // 3

#ifndef X11
    }
    else
    {
  
       ShowHelpPage = ShowHelpPageEGA;
       HelpPageQ = Demo ? 1 : 4;    // 6
    }
#endif    
}

void ChangeHelpPage(void)
{
    if (Ticks < NextHelpPageAt || ScreenBusy) return;

    HelpBusy = 1;
    {  RECT rc;
       rc.left = ScrollX; rc.right = ScrollX + ScrollW;
       rc.top = ScrollY;  rc.bottom = ScrollY + ScrollH;
       DrawSolidRect(&rc, BkgrColour);
    }

    HelpPageNo = (HelpPageNo + 1) % HelpPageQ;
    NextHelpPageAt = Ticks + HELP_INTERVAL;
    (*ShowHelpPage)();
    HelpBusy = 0;
}

static void ChangeBottomMode(int OldMode)
{
    if(OldMode == B_SCROLLER)
    { if (Demo) ShowScrollerFrame();
          else  ShowShadowFrame();
    }
    else
    if (OldMode == B_HELP)
         ShowScrollerFrame();

    RestartScroller = 1;
    PrepareScroller();
}

void ToggleBottomMode(void)
{   BOTTOM_MODE OldMode = BottomMode;

    if (BottomMode == B_HELP)
           while (HelpBusy && !TimerSuspended) pause();


    BottomMode = (BottomMode + 1) % 3;
    ChangeBottomMode(OldMode);

}

void MaintainHelp(void)
{   BOTTOM_MODE OldMode = BottomMode;

    if(BottomMode == B_HELP)
    {   NextHelpPageAt = Ticks;
        ChangeHelpPage();
        return;
    }

    BottomMode = B_HELP;
    ChangeBottomMode(OldMode);

}


int EnableJoystick(void)
{
/*   if (joyInit() > 0) return 0; */
   if (joySetParams() > 0) return 0;
   JoyEnabled = 1;
   return 1;
}

void DisableJoystick(void)
{
 /*   joyDeinit(); */
    JoyEnabled = 0;
}


char *GetPlayerName(void)
{  char *pname, *pname1;
   int pname_len;	
   
   // Getting player name
   pname = getenv("LOGNAME");
   if(!pname) pname  = getlogin(); 

   if (!pname)
      pname = "*** Unknown ***";
   else
   {   /* Try to get full nae */
       struct passwd   *p;
       p =  getpwnam(pname);
       if (p  && strlen(p->pw_gecos) > 0)
              pname = p->pw_gecos;
       
   }

   pname1 = strchr(pname, ',');
   if (pname1) pname_len = pname1 - pname;
        else   pname_len = strlen(pname);
        
   if (pname_len > 28) pname_len = 28;
   pname1 = malloc(pname_len+1);
   memcpy (pname1, pname, pname_len);
   pname1[pname_len] = '\0';
   return pname1;

}   
